@@ -13,80 +13,90 @@ use std::{
     rc::Rc,
 };
 
+use self::score::calculator::{FieldData, TermData};
+
 /**
  * Query Result.
 
  * `T` Document key.
 */
 #[derive(Debug, PartialEq)]
-pub struct QueryResult<I> {
+pub struct QueryResult<T> {
     /**
      * Document key.
      */
-    pub key: I,
+    pub key: T,
     /**
      * Result score.
      */
     pub score: f64,
 }
 
+pub fn max_score_merger(
+    score: &f64,
+    previous_score: Option<&f64>,
+    document_visited_for_term: bool,
+) -> f64 {
+    {
+        if let Some(p) = previous_score {
+            if document_visited_for_term {
+                f64::max(p.to_owned(), score.to_owned())
+            } else {
+                p + score
+            }
+        } else {
+            score.to_owned()
+        }
+    }
+}
+
 /**
 Performs a search with a simple free text query.
 All token separators work as a disjunction operator.
 Arguments
  * typeparam `T` Document key.
  * `index`.
- * `fields_boost` Fields boost factors.
- * `bm25k1` BM25 ranking function constant `k1`, controls non-linear term frequency normalization (saturation).
- * `bm25b` BM25 ranking function constant `b`, controls to what degree document length normalizes tf values.
+ * `query` Query string.
+ * `score_calculator` A struct that implements the ScoreCalculator trait to provide score calculations.
  * `tokenizer Tokenizer is a function that breaks a text into words, phrases, symbols, or other meaningful elements called tokens.
  * `filter` Filter is a function that processes tokens and returns terms, terms are used in Inverted Index to index documents.
+ * `fields_boost` Fields boost factors.
  * `remove`d Set of removed document keys.
- * `s` Query string.
+
 returns Array of QueryResult structs
- */
-pub fn query<T: Eq + Hash + Clone + Debug, S: ScoreCalculator<T>>(
+*/
+pub fn query<T: Eq + Hash + Clone + Debug, M, S: ScoreCalculator<T, M>>(
     index: &mut Index<T>,
-    fields_boost: &[f64],
-    /*  bm25k1: f64,
-    bm25b: f64,*/
-    score_calculator: &S,
+    query: &str,
+    score_calculator: &mut S,
     tokenizer: Tokenizer,
     filter: Filter,
+    fields_boost: &[f64],
     removed: Option<&HashSet<T>>,
-    s: &str,
 ) -> Vec<QueryResult<T>> {
     let docs = &index.docs;
     let fields = &index.fields;
-    let terms = tokenizer(&s);
+    let query_terms = tokenizer(&query);
     let mut scores: HashMap<T, f64> = HashMap::new();
-
-    for term_pre_filter in terms {
-        let term = filter(&term_pre_filter);
-        if !term.is_empty() {
-            let expanded_terms = expand_term(&index, &term);
-            let mut visited_documents: HashSet<T> = HashSet::new();
-            for e_term in expanded_terms {
-                let expansion_boost = {
-                    if e_term == term {
-                        1_f64
-                    } else {
-                        f64::ln(
-                            1_f64 + (1_f64 / (1_f64 + (e_term.len() as f64) - (term.len() as f64))),
-                        )
-                    }
-                };
-
-                let term_node_option = find_inverted_index_node(Rc::clone(&index.root), &e_term);
-
+    for query_term_pre_filter in &query_terms {
+        let query_term = filter(&query_term_pre_filter);
+        if !query_term.is_empty() {
+            let expanded_terms = expand_term(&index, &query_term);
+            let mut visited_documents_for_term: HashSet<T> = HashSet::new();
+            for query_term_expanded in expanded_terms {
+                let term_node_option =
+                    find_inverted_index_node(Rc::clone(&index.root), &query_term_expanded);
                 if let Some(term_node) = term_node_option {
-                    if let Some(term_node_option_first_doc) = &term_node.borrow().first_doc {
-                        let mut document_frequency = 0;
-                        let mut prev_pointer: Option<Rc<RefCell<DocumentPointer<T>>>> = None;
+                    let mut term_node_borrowed = term_node.borrow_mut();
+                    let mut new_first_doc = None;
+                    let mut assign_new_first_doc = false;
+                    let mut document_frequency = 0;
 
-                        let mut pointer_option = Some(Rc::clone(term_node_option_first_doc));
+                    if let Some(term_node_option_first_doc) = &mut term_node_borrowed.first_doc {
+                        let mut prev_pointer: Option<Rc<RefCell<DocumentPointer<T>>>> = None;
+                        let mut pointer_option = Some(Rc::clone(&term_node_option_first_doc));
                         while let Some(pointer) = pointer_option {
-                            if removed.is_some()
+                            if removed.is_some() // Cleanup old removed documents while searching. If vaccume after delete, this will have not effect
                                 && removed
                                     .unwrap()
                                     .contains(&pointer.borrow().details.borrow().key)
@@ -94,8 +104,9 @@ pub fn query<T: Eq + Hash + Clone + Debug, S: ScoreCalculator<T>>(
                                 if let Some(pp) = &prev_pointer {
                                     pp.borrow_mut().next = pointer.borrow().next.clone();
                                 } else {
-                                    term_node.borrow_mut().first_doc =
-                                        (&pointer.borrow().next).clone();
+                                    new_first_doc = (&pointer.borrow().next).clone();
+                                    assign_new_first_doc = true;
+                                    //  term_node_borrowed.first_doc = (&pointer.borrow().next).clone();
                                 }
                             } else {
                                 prev_pointer = Some(Rc::clone(&pointer));
@@ -103,16 +114,26 @@ pub fn query<T: Eq + Hash + Clone + Debug, S: ScoreCalculator<T>>(
                             }
                             pointer_option = pointer.borrow().next.clone();
                         }
+                    }
 
+                    if assign_new_first_doc {
+                        term_node_borrowed.first_doc = new_first_doc;
+                    }
+
+                    if let Some(term_node_option_first_doc) = &mut term_node_borrowed.first_doc {
                         if document_frequency > 0 {
-                            // calculating BM25 idf
-                            let idf = f64::ln(
-                                1_f64
-                                    + ((docs.len() - document_frequency) as f64 + 0.5)
-                                        / (document_frequency as f64 + 0.5),
+                            let term_expansion_data = TermData {
+                                all_query_terms: &query_terms,
+                                query_term: &query_term,
+                                query_term_expanded: &query_term_expanded,
+                            };
+                            let pre_calculations = &score_calculator.before_each(
+                                &term_expansion_data,
+                                document_frequency,
+                                docs,
                             );
 
-                            let mut pointer = Some(Rc::clone(term_node_option_first_doc));
+                            let mut pointer = Some(Rc::clone(&term_node_option_first_doc));
                             while let Some(p) = pointer {
                                 let pointer_borrowed = p.borrow();
                                 let field_lengths = &pointer_borrowed.details.borrow().field_length;
@@ -121,29 +142,25 @@ pub fn query<T: Eq + Hash + Clone + Debug, S: ScoreCalculator<T>>(
                                         .unwrap()
                                         .contains(&pointer_borrowed.details.borrow().key)
                                 {
-                                    let score = score_calculator.score(
+                                    let score = &score_calculator.score(
+                                        pre_calculations.as_ref(),
                                         p.borrow(),
-                                        idf,
-                                        field_lengths,
-                                        fields_boost,
-                                        expansion_boost,
-                                        fields,
+                                        &FieldData {
+                                            field_lengths,
+                                            fields_boost,
+                                            fields,
+                                        },
+                                        &term_expansion_data,
                                     );
-                                    if score > 0_f64 {
+                                    if let Some(s) = score {
                                         let key = &pointer_borrowed.details.borrow().key;
-                                        let new_score = {
-                                            if let Some(prev_score) = scores.get(&key) {
-                                                if visited_documents.contains(&key) {
-                                                    f64::max(prev_score.to_owned(), score)
-                                                } else {
-                                                    prev_score + score
-                                                }
-                                            } else {
-                                                score
-                                            }
-                                        };
+                                        let new_score = max_score_merger(
+                                            s,
+                                            scores.get(&key),
+                                            visited_documents_for_term.contains(&key),
+                                        );
                                         scores.insert(key.to_owned(), new_score);
-                                        visited_documents.insert(key.to_owned());
+                                        visited_documents_for_term.insert(key.to_owned());
                                     }
                                 }
                                 pointer = pointer_borrowed.next.clone();
@@ -161,6 +178,9 @@ pub fn query<T: Eq + Hash + Clone + Debug, S: ScoreCalculator<T>>(
     }
 
     result.sort_by(|a, b| b.score.partial_cmp(&a.score).unwrap());
+
+    score_calculator.finalize(&mut result);
+
     result
 }
 
@@ -242,8 +262,6 @@ mod tests {
     }
 
     pub mod query {
-        use crate::query::score;
-
         use super::*;
 
         #[test]
@@ -273,12 +291,12 @@ mod tests {
             }
             let result = query(
                 &mut idx,
-                &vec![1., 1.],
-                &score::default::bm25::default(),
+                &"a".to_string(),
+                &mut crate::query::score::default::bm25::new(),
                 tokenizer,
                 filter,
+                &vec![1., 1.],
                 None,
-                &"a".to_string(),
             );
             assert_eq!(result.len(), 1);
             assert_eq!(
@@ -317,12 +335,12 @@ mod tests {
 
             let result = query(
                 &mut idx,
-                &vec![1., 1.],
-                &score::default::bm25::default(),
+                &"c".to_string(),
+                &mut crate::query::score::default::bm25::new(),
                 tokenizer,
                 filter,
+                &vec![1., 1.],
                 None,
-                &"c".to_string(),
             );
             assert_eq!(result.len(), 2);
             assert_eq!(
@@ -373,12 +391,12 @@ mod tests {
 
             let result = query(
                 &mut idx,
-                &vec![1., 1.],
-                &score::default::bm25::default(),
+                &"h".to_string(),
+                &mut crate::query::score::default::bm25::new(),
                 tokenizer,
                 filter,
+                &vec![1., 1.],
                 None,
-                &"h".to_string(),
             );
             assert_eq!(result.len(), 1);
             assert_eq!(
@@ -423,12 +441,12 @@ mod tests {
             }
             let result = query(
                 &mut idx,
-                &vec![1., 1.],
-                &score::default::bm25::default(),
+                &"a".to_string(),
+                &mut crate::query::score::default::bm25::new(),
                 tokenizer,
                 custom_filter,
+                &vec![1., 1.],
                 None,
-                &"a".to_string(),
             );
             assert_eq!(result.len(), 0);
         }
@@ -462,12 +480,12 @@ mod tests {
 
             let result = query(
                 &mut idx,
-                &vec![1., 1.],
-                &score::default::bm25::default(),
+                &"a d".to_string(),
+                &mut crate::query::score::default::bm25::new(),
                 tokenizer,
                 filter,
+                &vec![1., 1.],
                 None,
-                &"a d".to_string(),
             );
             assert_eq!(result.len(), 2);
             assert_eq!(
